// TODO: worker run 
import Color from './Color'
import Canvas from './Canvas'
import StackBlur from './blur/StackBlur'
import Matrix from './Matrix'

function weight(arr, num = 1) {
    return arr.map(i => {
        return i * num;
    })
}

function repeat (value, num) {
    let arr = new Array(num)
    for(let i = 0; i < num; i++) {
        arr[i] = value 
    }
    return arr; 
}

function colorMatrix(pixels, i, matrix) {
    var r = pixels[i], g = pixels[i + 1], b = pixels[i + 2], a = pixels[i + 3];

    pixels[i] = matrix[0] * r + matrix[1] * g + matrix[2] * b + matrix[3] * a;
    pixels[i+1] = matrix[4] * r + matrix[5] * g + matrix[6] * b + matrix[7] * a;
    pixels[i+2] = matrix[8] * r + matrix[9] * g + matrix[10] * b + matrix[11] * a;
    pixels[i+3] = matrix[12] * r + matrix[13] * g + matrix[14] * b + matrix[15] * a;

}

function makeFilter(filter) {

    if (typeof filter == 'function') {
        return filter;
    }

    if (typeof filter == 'string') {
        filter = [filter];
    }

    const filterName = filter.shift();

    if (typeof filterName == 'function') {
        return filterName;
    }

    const params = filter;

    const filterFunction = F[filterName];

    if (!filterFunction) {
        throw new Error(`${filterName} is not filter. please check filter name.`)
    }
    return filterFunction.apply(filterFunction, params);
}

function each(len, callback) {
    for (var i = 0, xyIndex = 0; i < len; i += 4, xyIndex++) {
        callback(i, xyIndex);
    }
}

function eachXY(len, width, callback) {
    for (var i = 0, xyIndex = 0; i < len; i += 4, xyIndex++) {
        callback(i, xyIndex % width, Math.floor(xyIndex / width) );
    }
}

function createRandRange(min, max, count) {
    var result = [];

    for (var i = 1; i <= count; i++) {
        var num = Math.random() * (max - min) + min;
        var sign = (Math.floor(Math.random() * 10) % 2 == 0) ? -1 : 1;
        result.push(sign * num);
    }

    result.sort();

    const centerIndex = Math.floor(count / 2);
    var a = result[centerIndex];
    result[centerIndex] = result[0];
    result[0] = a;

    return result;
}

function createRandomCount() {
    return [3 * 3, 4 * 4, 5 * 5, 6 * 6, 7 * 7, 8 * 8, 9 * 9, 10 * 10].sort(function (a, b) {
        return 0.5 - Math.random();
    })[0];
}

function createBitmap(length, width, height) {
    return { pixels: new Uint8ClampedArray(length), width, height }
}

function getBitmap(bitmap, area) {
    return Canvas.getBitmap(bitmap, area);
}

function putBitmap(bitmap, subBitmap, area) {
    return Canvas.putBitmap(bitmap, subBitmap, area);
}

function parseParamNumber (param) {
    if (typeof param === 'string') {
        param = param.replace(/deg/, '')
        param = param.replace(/px/, '')
    }
    return +param 
} 

let F = {};
let ImageFilter = F

const filter_regexp = /(([\w_\-]+)(\(([^\)]*)\))?)+/gi;
const filter_split = ' '


const pack = F.pack = function pack(callback) {
    return function (bitmap) {
        each(bitmap.pixels.length, (i, xyIndex) => {
            callback(bitmap.pixels, i, xyIndex)
        })
        return bitmap;
    }
}


const ColorListIndex = [0, 1, 2, 3]

const swapColor = F.swapColor = function swapColor (pixels, startIndex, endIndex) {

    ColorListIndex.forEach(i => {
        var temp = pixels[startIndex + i]
        pixels[startIndex + i] = pixels[endIndex + i]
        pixels[endIndex + i] = temp
    })
}

const packXY = F.packXY = function packXY(callback) {
    return function (bitmap) {
        eachXY(bitmap.pixels.length, bitmap.width, (i, x, y) => {
            callback(bitmap.pixels, i, x, y)
        })
        return bitmap;
    }
}

F.matches = function (str) {
    var ret = Color.convertMatches(str)
    const matches = ret.str.match(filter_regexp);
    let result = [];

    if (!matches) {
        return result;
    }

    result = matches.map((it) => {
        return { filter: it, origin: Color.reverseMatches(it, ret.matches) }
    })

    var pos = { next: 0 }
    result = result.map(item => {

        const startIndex = str.indexOf(item.origin, pos.next);

        item.startIndex = startIndex;
        item.endIndex = startIndex + item.origin.length;

        item.arr = this.parseFilter(item.origin) 

        pos.next = item.endIndex;

        return item 
    }).filter(it => {
        if (!it.arr.length) return false 
        return !!F[it.arr[0]]
    })

    return result;
}

/**
 * Filter Parser 
 * 
 * F.parseFilter('blur(30)') == ['blue', '30']
 * F.parseFilter('gradient(white, black, 3)') == ['gradient', 'white', 'black', '3']
 * 
 * @param {String} filterString 
 */
F.parseFilter = function (filterString) {

    var ret = Color.convertMatches(filterString)
    const matches = ret.str.match(filter_regexp);

    if (!matches[0]) {
        return []
    }

    var arr = matches[0].split('(')

    var filterName = arr.shift()
    var filterParams = [] 

    if (!F[filterName.toLowerCase()]) {
        return []
    }

    if (arr.length) {
        filterParams = arr.shift().split(')')[0].split(',').map(f => {
            return Color.reverseMatches(f, ret.matches)
        })    
    }
    
    var result = [filterName, ...filterParams].map(Color.trim)
    
    return result 
}

F.filter = function (str) {
    return F.merge(F.matches(str).map(it => {
        return it.arr 
    }))
}

/** 
 * 
 * multiply filters
 * 
 * ImageFilter.multi('blur', 'grayscale', 'sharpen', ['blur', 3], function (bitmap) {  return bitmap });
 * 
 */
F.multi = function (...filters) {
    filters = filters.map(f => {
        return makeFilter(f);
    })

    return function (bitmap) {
        return filters.reduce((bitmap, f) => {
            return f(bitmap);
        }, bitmap)
    }
}


F.merge = function (filters) {
    return F.multi(...filters);
}

/**
 * apply filter into special area
 * 
 * F.partial({x,y,width,height}, filter, filter, filter )
 * F.partial({x,y,width,height}, 'filter' )
 * 
 * @param {{x, y, width, height}} area 
 * @param {*} filters 
 */
F.partial = function (area, ...filters) {
    var allFilter = null 
    if (filters.length == 1 && typeof filters[0] === 'string') {
        allFilter = F.filter(filters[0])
    } else {
        allFilter = F.merge(filters)
    } 

    return (bitmap) => {
        return putBitmap(bitmap, allFilter(getBitmap(bitmap, area)), area);
    }
}

F.counter = function (filter, count = 1) {
    var filters = [];

    for (var i = 0; i < count; i++) {
        filters.push(filter);
    }

    return F.multi(...filters);
}

// Image manupulate 
F.resize = function (dstWidth, dstHeight) {
    return function (bitmap) {
        
        var c = Canvas.drawPixels(bitmap);
        var context = c.getContext('2d');

        c.width = dstWidth;
        c.height = dstHeight;

        return {
            pixels: new Uint8ClampedArray(context.getImageData(0, 0, dstWidth, dstHeight).data),
            width: dstWidth,
            height: dstHeight
        }
    }
}

F.crop = function (startX = 0, startY = 0, width, height) {

    const newBitmap = createBitmap(width * height * 4, width, height)

    return function (bitmap) {
        for (var y = startY, realY = 0; y < height; y++, realY++) {
            for (var x = startX, realX = 0; x < width; x++, realX++) {
                newBitmap.pixels[realY * width * realX] = bitmap.pixels[y * width * x]
            }
        }

        return newBitmap;
    }
}


F.flipH = function flipH () {
    return function (bitmap) {

        const width = bitmap.width
        const height = bitmap.height 
        const isCenter = width % 2 == 1 ? 1 : 0 

        const halfWidth = isCenter ? Math.floor(width / 2) : width / 2 ;

        for (var y = 0; y < height; y++) {
            for (var x = 0; x < halfWidth; x++) {

                var startIndex = (y * width + x) * 4 
                var endIndex = (y * width +  (width -1 - x) ) * 4 
                swapColor(bitmap.pixels, startIndex, endIndex)

            }
        }

        return bitmap;
    }
}

F.flipV = function flipV () {
    return function (bitmap) {

        const width = bitmap.width
        const height = bitmap.height 
        const isCenter = height % 2 == 1 ? 1 : 0 

        const halfHeight = isCenter ? Math.floor(height / 2) : height / 2 ;

        for (var y = 0; y < halfHeight; y++) {
            for (var x = 0; x < width; x++) {

                var startIndex = (y * width + x) * 4 
                var endIndex = ((height -1 - y) * width + x ) * 4 
                swapColor(bitmap.pixels, startIndex, endIndex)

            }
        }

        return bitmap;
    }
}

F.radian = function (degree) {
    return Matrix.CONSTANT.radian(degree)
}

F['rotate-degree'] = F.rotateDegree = function (angle, cx = 'center', cy = 'center') {
    // const r = F.radian(angle)

    return function (bitmap) {
        var newBitmap = createBitmap(bitmap.pixels.length, bitmap.width, bitmap.height)
        const width = bitmap.width 
        const height = bitmap.height 

        if (cx == 'center') {
            cx = Math.floor(width / 2); 
        }

        if (cy == 'center') {
            cy = Math.floor(height/ 2);
        }

        const translateMatrix = Matrix.CONSTANT.translate(-cx, -cy)
        const translateMatrix2 = Matrix.CONSTANT.translate(cx, cy)
        const shear1Matrix = Matrix.CONSTANT.shear1(angle)
        const shear2Matrix = Matrix.CONSTANT.shear2(angle)

        return packXY((pixels, i, x, y) => {
            // console.log(x, y, i)
            let arr = Matrix.multiply(translateMatrix, [x, y, 1])
            
            arr = Matrix.multiply(shear1Matrix, arr).map(Math.round)
            arr = Matrix.multiply(shear2Matrix, arr).map(Math.round)
            arr = Matrix.multiply(shear1Matrix, arr).map(Math.round)
            arr = Matrix.multiply(translateMatrix2, arr)
 
            const [x1, y1] = arr

            if (x1 < 0) return;   
            if (y1 < 0) return; 
            if (x1 > width-1) return;
            if (y1 > height-1) return; 

            var endIndex = (y1 * width + x1) * 4 

            pixels[endIndex] = bitmap.pixels[i]
            pixels[endIndex+1] = bitmap.pixels[i+1]
            pixels[endIndex+2] = bitmap.pixels[i+2]
            pixels[endIndex+3] = bitmap.pixels[i+3]

        })(newBitmap)
    }
} 


F.rotate = function rotate (degree = 0) {
    degree = parseParamNumber(degree)     
    degree = degree % 360
    return function (bitmap) {

        if (degree == 0) return bitmap

        if (degree == 90 || degree == 270) {
            var newBitmap = createBitmap(bitmap.pixels.length, bitmap.height, bitmap.width)
        } else if (degree == 180) {
            var newBitmap = createBitmap(bitmap.pixels.length, bitmap.width, bitmap.height)
        } else {
            return F.rotateDegree(degree)(bitmap)
        }

        const width = bitmap.width 
        const height = bitmap.height 

        packXY((pixels, i, x, y) => {

            if (degree == 90) {
                var endIndex = (x * newBitmap.width + (newBitmap.width -1 - y) ) * 4 
            } else if (degree == 270) {
                var endIndex = ( (newBitmap.height -1 -x) * newBitmap.width + y ) * 4 
            } else if (degree == 180) {
                var endIndex = ((newBitmap.height -1 -y) * newBitmap.width + (newBitmap.width -1 -x)) * 4
            }

            // console.log(startIndex, endIndex)

            newBitmap.pixels[endIndex] = bitmap.pixels[i]
            newBitmap.pixels[endIndex+1] = bitmap.pixels[i+1]
            newBitmap.pixels[endIndex+2] = bitmap.pixels[i+2]
            newBitmap.pixels[endIndex+3] = bitmap.pixels[i+3]

        })(bitmap)

        return newBitmap
    }
}

// Pixel based 


F.grayscale = function (amount) { 
    amount = parseParamNumber(amount)          
    let C = amount / 100;

    if (C > 1) C = 1; 
    
    return pack((pixels, i) => {

        colorMatrix(pixels, i, [
            (0.2126 + 0.7874 * (1 - C)), (0.7152 - 0.7152 * (1 - C)), (0.0722 - 0.0722 * (1 - C)), 0,
            (0.2126 - 0.2126 * (1 - C)), (0.7152 + 0.2848 * (1 - C)), (0.0722 - 0.0722 * (1 - C)), 0,
            (0.2126 - 0.2126 * (1 - C)), (0.7152 - 0.7152 * (1 - C)), (0.0722 + 0.9278 * (1 - C)), 0,
            0, 0, 0, 1
        ])
    });
}

/*
 * @param {Number} amount   0..360  
 */
F.hue = function (amount = 360) {
    amount = parseParamNumber(amount)          
    return pack((pixels, i) => {
        var r = pixels[i], g = pixels[i + 1], b = pixels[i + 2];

        var hsv = Color.RGBtoHSV(r, g, b);

        // 0 ~ 360 
        var h = hsv.h;
        h += Math.abs(amount)
        h = h % 360
        hsv.h = h

        var rgb = Color.HSVtoRGB(hsv);

        pixels[i] = rgb.r;
        pixels[i + 1] = rgb.g;
        pixels[i + 2] = rgb.b;

    })
}


F.shade = function (r = 1, g = 1, b = 1) {
    r = parseParamNumber(r)        
    g = parseParamNumber(g)        
    b = parseParamNumber(b)        
    return pack((pixels, i) => {
        pixels[i] *= r;
        pixels[i + 1] *= g;
        pixels[i + 2] *= b;
    })

}

F.bitonal = function (darkColor, lightColor, threshold = 100) {
    darkColor = Color.parse(darkColor);
    lightColor = Color.parse(lightColor);
    return pack((pixels, i) => {

        if (pixels[i] + pixels[i + 1] + pixels[i + 2] <= threshold) {
            pixels[i] = darkColor.r;
            pixels[i + 1] = darkColor.g;
            pixels[i + 2] = darkColor.b;
        } else {
            pixels[i] = lightColor.r;
            pixels[i + 1] = lightColor.g;
            pixels[i + 2] = lightColor.b;
        }
    })
}


/**
 * F.gradient('red', 'blue', 'yellow', 'white', 10)
 * F.gradient('red, blue, yellow, white, 10')
 */
F.duotone = function () {
    // 전체 매개변수 기준으로 파싱 
    // 색이 아닌 것 기준으로 scale 변수로 인식 

    let params = [...arguments];

    if (params.length === 1 && typeof params[0] === 'string') {
        params = Color.convertMatchesArray(params[0])
    } 

    params = params.map(arg => {
        const res = Color.matches(arg)

        if (!res.length) {
            return { type: 'scale', value : arg }
        }

        return { type: 'param', value : arg }
    })

    let scale = params.filter(it => { return it.type == 'scale' })[0]
    scale = scale ? +scale.value : 256

    params = params.filter(it => { return it.type == 'param' }).map( it => {
        return it.value 
    }).join(',')

    let colors = Color.gradient(params, scale).map(c => { return Color.parse(c) })

    return pack((pixels, i) => {
        const colorIndex = F.clamp(Color.brightness(pixels[i] , pixels[i + 1] , pixels[i + 2]))
        const newColorIndex = F.clamp(Math.floor(colorIndex * (scale / 256)))
        const color = colors[newColorIndex]

        pixels[i] = color.r;
        pixels[i + 1] = color.g;
        pixels[i + 2] = color.b;
        pixels[i + 3] = F.clamp(Math.floor(color.a * 256));
    })
}

F.gradient = F.duotone;

F.tint = function (redTint = 1, greenTint = 1, blueTint = 1) {
    redTint = parseParamNumber(redTint)       
    greenTint = parseParamNumber(greenTint)       
    blueTint = parseParamNumber(blueTint)       
    return pack((pixels, i) => {
        pixels[i] += (255 - pixels[i]) * redTint;
        pixels[i + 1] += (255 - pixels[i + 1]) * greenTint;
        pixels[i + 2] += (255 - pixels[i + 2]) * blueTint;
    })

}

F.clamp = function (num) {
    return Math.min(255, num)
}
/**
 * 
 * @param {*} amount   min = -128, max = 128 
 */
F.contrast = function (amount = 0) {
    amount = parseParamNumber(amount)       
    const C = Math.max((128 + amount) / 128, 0);

    return pack((pixels, i) => {
        pixels[i] = pixels[i] * C
        pixels[i+1] = pixels[i+1] * C
        pixels[i+2] = pixels[i+2] * C
    })
}

F.invert = function (amount = 100) {
    amount = parseParamNumber(amount)    
    const C = amount / 100; 

    return pack((pixels, i) => {

        pixels[i] = (255 - pixels[i]) * C ;
        pixels[i + 1] = (255 - pixels[i + 1]) * C;
        pixels[i + 2] = (255 - pixels[i + 2]) * C;

    })
}

F.opacity = function (amount = 100) {
    amount = parseParamNumber(amount)   
    const C = amount / 100; 

    return pack((pixels, i) => {
        pixels[i + 3] *= C;
    })
}

/**
 * change the relative darkness of (a part of an image) by overexposure to light.
 * @param {*} r 
 * @param {*} g 
 * @param {*} b 
 */
F.solarize = function (r, g, b) {
    r = parseParamNumber(r)    
    g = parseParamNumber(g)    
    b = parseParamNumber(b)    
    return pack((pixels, i) => {
        if (pixels[i] < r) pixels[i] = 255 - pixels[i];
        if (pixels[i + 1] < g) pixels[i + 1] = 255 - pixels[i + 1];
        if (pixels[i + 2] < b) pixels[i + 2] = 255 - pixels[i + 2];
    })

}

/*
 * @param {Number} amount  0..100 
 */
F.sepia = function (amount = 100) {
    amount = parseParamNumber(amount)    
    let C = amount / 100;
    if (C > 1) C = 1; 

    return pack((pixels, i) => {

        colorMatrix(pixels, i, [
            (0.393 + 0.607 * (1 - C)), (0.769 - 0.769 * (1 - C)), (0.189 - 0.189 * (1 - C)), 0,
            (0.349 - 0.349 * (1 - C)), (0.686 + 0.314 * (1 - C)), (0.168 - 0.168 * (1 - C)), 0,
            (0.272 - 0.272 * (1 - C)), (0.534 - 0.534 * (1 - C)), (0.131 + 0.869 * (1 - C)), 0,
            0, 0, 0, 1
        ])
    })
}

F.gamma = function (amount = 1) {
    amount = parseParamNumber(amount)    
    return pack((pixels, i) => {
        pixels[i] = Math.pow(pixels[i] / 255, amount) * 255
        pixels[i+1] = Math.pow(pixels[i+1] / 255, amount) * 255
        pixels[i+2] = Math.pow(pixels[i+2] / 255, amount) * 255
    })
}

/**
 * 
 * @param {Number} amount 1..100
 */
F.noise = function (amount = 1) {
    amount = parseParamNumber(amount)    
    return pack((pixels, i) => {
        const C = Math.abs(amount) * 5
        const min = -C
        const max = C 
        const noiseValue = Math.round(min + (Math.random() * (max - min)))
        pixels[i] += noiseValue 
        pixels[i+1] += noiseValue 
        pixels[i+2] += noiseValue 
    })
}

/**
 * 
 * @param {Number} amount from 0 to 100 
 */
F.clip = function (amount = 0) {
    amount = parseParamNumber(amount)    
    const C = Math.abs(amount) * 2.55

    return pack((pixels, i) => {

        for(var start = i, end = i + 2; start <= end; start++) {
            if (pixels[start] > 255 - C) {
                pixels[start] = 255 
            } else if (pixels[start] < C) {
                pixels[start] = 0 
            }            
        }

    })
}

/*
 * @param {Number} amount  -100..100  ,  value < 0  is darken, value > 0 is brighten 
 */
F.brightness = function (amount = 1) {
    amount = parseParamNumber(amount)    
    const C = Math.floor(255 * (amount / 100));

    return pack((pixels, i) => {
        pixels[i] += C;
        pixels[i + 1] += C;
        pixels[i + 2] += C;
    })
}

/*
 * @param {Number} amount  -100..100 
 */
F.saturation = function (amount = 100) {
    amount = parseParamNumber(amount)    
    const C = amount / 100 
    const L = 1 - Math.abs(C);
    return pack((pixels, i) => {

        colorMatrix(pixels, i, [
            L, 0, 0, 0,
            0, L, 0, 0,
            0, 0, L, 0,
            0, 0, 0, L
        ]);
    })

}


/*
 * @param {Number} amount  0..100 
 */
F.threshold = function (scale = 200, amount = 100) {
    return F.thresholdColor(scale, amount, false)
}

F['threshold-color'] = F.thresholdColor = function (scale = 200, amount = 100, hasColor = true) {
    scale = parseParamNumber(scale)    
    amount = parseParamNumber(amount)    
    const C = amount / 100;
    return pack((pixels, i) => {
        var v = (C * Color.brightness(pixels[i], pixels[i + 1], pixels[i + 2]) ) >= scale ? 255 : 0;

        if (hasColor) {

            if (v == 0) {
                pixels[i] = pixels[i + 1] = pixels[i + 2] = 0
            }
            
        } else {
            pixels[i] = pixels[i + 1] = pixels[i + 2] = Math.round(v)
        }
        
    })
}

// Matrix based 

F.convolution = function (weights, opaque = true) {
    return function ({ pixels, width, height }) {
        const side = Math.round(Math.sqrt(weights.length));
        const halfSide = Math.floor(side / 2);

        var w = width;
        var h = height;
        var sw = w;
        var sh = h;
        let dst = new Uint8ClampedArray(pixels.length);
        const alphaFac = opaque ? 1 : 0;

        for (var y = 0; y < h; y++) {
            for (var x = 0; x < w; x++) {
                const sy = y;
                const sx = x;
                const dstIndex = (y * w + x) * 4;

                var r = 0, g = 0, b = 0, a = 0;
                for (var cy = 0; cy < side; cy++) {
                    for (var cx = 0; cx < side; cx++) {

                        const scy = sy + cy - halfSide;
                        const scx = sx + cx - halfSide;

                        if (scy >= 0 && scy < sh && scx >= 0 && scx < sw) {
                            var srcIndex = (scy * sw + scx) * 4;
                            var wt = weights[cy * side + cx];
                            r += pixels[srcIndex] * wt;
                            g += pixels[srcIndex + 1] * wt;
                            b += pixels[srcIndex + 2] * wt;
                            a += pixels[srcIndex + 3] * wt;   // weight 를 곱한 값을 계속 더한다. 
                        }
                    }
                }

                dst[dstIndex] = r;
                dst[dstIndex + 1] = g;
                dst[dstIndex + 2] = b;
                dst[dstIndex + 3] = a + alphaFac * (255 - a);
            }
        }

        return { pixels: dst, width: sw, height: sh };
    }
}

F.identity = function () {
    return F.convolution([
        0, 0, 0,
        0, 1, 0,
        0, 0, 0
    ]);
}

F.random = function (amount = 10, count = createRandomCount()) {
    amount = parseParamNumber(amount)    
    return function (pixels, width, height) {
        var rand = createRandRange(-1, 5, count);
        return F.convolution(rand)(pixels, width, height);
    }
}


F.grayscale2 = function (amount = 100) {
    amount = parseParamNumber(amount)    
    return F.convolution(weight([
        0.3, 0.3, 0.3, 0, 0,
        0.59, 0.59, 0.59, 0, 0,
        0.11, 0.11, 0.11, 0, 0,
        0, 0, 0, 0, 0,
        0, 0, 0, 0, 0
    ], amount / 100));
}

F.sepia2 = function (amount = 100) {
    amount = parseParamNumber(amount)    
    return F.convolution(weight([
        0.393, 0.349, 0.272, 0, 0,
        0.769, 0.686, 0.534, 0, 0,
        0.189, 0.168, 0.131, 0, 0,
        0, 0, 0, 0, 0,
        0, 0, 0, 0, 0
    ], amount / 100));
}

F.negative = function (amount = 100) {
    amount = parseParamNumber(amount)    
    return F.convolution(weight([
        -1, 0, 0, 0, 0,
        0, -1, 0, 0, 0,
        0, 0, -1, 0, 0,
        0, 0, 0, 1, 0,
        1, 1, 1, 1, 1
    ], amount / 100));
}

F.sharpen = function (amount = 100) {
    amount = parseParamNumber(amount)    
    return F.convolution(weight([
        0, -1, 0,
        -1, 5, -1,
        0, -1, 0
    ], amount / 100));
}

F['motion-blur'] = F.motionBlur = function () {
    return F.convolution(weight([
        1, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 1, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 1, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 1, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 1, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 1, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 1, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 1, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 1,
    ], 1 / 9));
}

F['motion-blur-2'] = F.motionBlur2 = function () {
    return F.convolution(weight([
        1, 0, 0, 0, 0, 0, 0, 0, 1,
        0, 1, 0, 0, 0, 0, 0, 1, 0,
        0, 0, 1, 0, 0, 0, 1, 0, 0,
        0, 0, 0, 1, 0, 1, 0, 0, 0,
        0, 0, 0, 0, 1, 0, 0, 0, 0,
        0, 0, 0, 1, 0, 1, 0, 0, 0,
        0, 0, 1, 0, 0, 0, 1, 0, 0,
        0, 1, 0, 0, 0, 0, 0, 1, 0,
        1, 0, 0, 0, 0, 0, 0, 0, 1,
    ], 1 / 9));
}

F['motion-blur-3'] = F.motionBlur3 = function () {
    return F.convolution(weight([
        1, 0, 0, 0, 1, 0, 0, 0, 1,
        0, 1, 0, 0, 1, 0, 0, 1, 0,
        0, 0, 1, 0, 1, 0, 1, 0, 0,
        0, 0, 0, 1, 1, 1, 0, 0, 0,
        1, 1, 1, 1, 1, 1, 1, 1, 1,
        0, 0, 0, 1, 1, 1, 0, 0, 0,
        0, 0, 1, 0, 1, 0, 1, 0, 0,
        0, 1, 0, 0, 1, 0, 0, 1, 0,
        1, 0, 0, 0, 1, 0, 0, 0, 1,
    ], 1 / 9));
}

function createBlurMatrix (amount = 3) {
    const count = Math.pow(amount, 2)
    const value = 1/count
    return repeat (value, count)
}

/**
 * 
 * @param {Number} amount   from 3 to 100 
 */
F.blur = function (amount = 3, hasAlphaChannel = true) {
    amount = parseParamNumber(amount)

    return F.convolution(createBlurMatrix(amount))
}

F['stack-blur'] = F.stackBlur = function (radius = 10, hasAlphaChannel = true) {
    radius = parseParamNumber(radius)

    return function (bitmap) {
        return StackBlur(bitmap, radius, hasAlphaChannel )
    }
}

F['gaussian-blur'] = F.gaussianBlur = function (amount = 100) {
    amount = parseParamNumber(amount)    
    const C = amount / 100; 

    return F.convolution(weight([
        1, 2, 1,
        2, 4, 2,
        1, 2, 1
    ], (1/16) * C ));
}

F['gaussian-blur-5x'] = F.gaussianBlur5x = function (amount = 100) {
    amount = parseParamNumber(amount)    
    const C = amount / 100;
    return F.convolution(weight([
        1, 4, 6, 4, 1,
        4, 16, 24, 16, 4,
        6, 24, 36, 24, 6,
        4, 16, 24, 16, 4,
        1, 4, 6, 4, 1
    ], (1/256) * C ));
}

F['unsharp-masking'] = F.unsharpMasking = function (amount = 256) {
    amount = parseParamNumber(amount)
    return F.convolution(weight([
        1, 4, 6, 4, 1,
        4, 16, 24, 16, 4,
        6, 24, -476, 24, 6,
        4, 16, 24, 16, 4,
        1, 4, 6, 4, 1
    ], -1 / amount));
}

F.transparency = function (amount = 100) {
    amount = parseParamNumber(amount)
    return F.convolution(weight([
        1, 0, 0, 0, 0,
        0, 1, 0, 0, 0,
        0, 0, 1, 0, 0,
        0, 0, 0, 0.3, 0,
        0, 0, 0, 0, 1,
    ], amount / 100));
}

F.laplacian = function (amount = 100) {
    amount = parseParamNumber(amount)
    return F.convolution(weight([
        -1, -1, -1,
        -1, 8, -1,
        -1, -1, -1
    ], amount / 100));
}

F.laplacian.grayscale = function (amount = 100) {
    return F.filter(`grayscale laplacian(${amount}`)
}

F.laplacian5x = function (amount = 100) {
    amount = parseParamNumber(amount)
    return F.convolution(weight([
        -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1,
        -1, -1, 24, -1, -1,
        -1, -1, -1, -1, -1,
        -1, -1, -1, -1, -1
    ], amount / 100));
}

F.laplacian5x.grayscale = function () {
    return F.filter('grayscale laplacian5x');
}

F['kirsch-horizontal'] = F.kirschHorizontal = function (count = 1) {
    count = parseParamNumber(count)    
    return F.convolution([
        5, 5, 5,
        -3, 0, -3,
        -3, -3, -3
    ]);
}

F['kirsch-vertical'] = F.kirschVertical = function (count = 1) {
    count = parseParamNumber(count)    
    return F.convolution([
        5, -3, -3,
        5, 0, -3,
        5, -3, -3
    ]);
}

F.kirsch = function () {
    return F.filter('kirsch-horizontal kirsch-vertical');
}

F.kirsch.grayscale = function () {
    return F.filter('grayscale kirsch');
}

F['sobel-horizontal'] = F.sobelHorizontal = function () {
    return F.convolution([
        -1, -2, -1,
        0, 0, 0,
        1, 2, 1
    ]);
}

F['sobel-vertical'] = F.sobelVertical = function () {
    return F.convolution([
        -1, 0, 1,
        -2, 0, 2,
        -1, 0, 1
    ]);
}

F.sobel = function () {
    return F.filter('sobel-horizontal sobel-vertical');
}

F.sobel.grayscale = function () {
    return F.filter('grayscale sobel');
}

/*
 * carve, mold, or stamp a design on (a surface) so that it stands out in relief.
 * 
 * @param {Number} amount   0.0 .. 4.0 
 */
F.emboss = function (amount = 4) {
    amount = parseParamNumber(amount)    
    return F.convolution([
        amount * (-2.0), -amount, 0.0,
        -amount, 1.0, amount,
        0.0, amount, amount * 2.0,
    ]);
}

/**
 * multi filter 
 */

F.vintage = function () {
    return F.filter(`brightness(15) saturation(-20) gamma(1.8)`)
}

export default ImageFilter